/*
 * Copyright 2015 - 2018 gtalent2@gmail.com
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#include <ox/std/byteswap.hpp>
#include <ox/std/string.hpp>
#include <ox/std/vector.hpp>
#include <ox/std/types.hpp>
#include "err.hpp"
#include "optype.hpp"
#include "presencemask.hpp"

namespace ox {

class MetalClawWriter {

	private:
		FieldPresenseMask m_fieldPresence;
		const char *m_typeName = nullptr;
		int m_fields = 0;
		int m_field = 0;
		std::size_t m_buffIt = 0;
		std::size_t m_buffLen = 0;
		uint8_t *m_buff = nullptr;

	public:
		MetalClawWriter(uint8_t *buff, std::size_t buffLen);

		~MetalClawWriter() noexcept;

		int op(const char*, int8_t *val);
		int op(const char*, int16_t *val);
		int op(const char*, int32_t *val);
		int op(const char*, int64_t *val);

		int op(const char*, uint8_t *val);
		int op(const char*, uint16_t *val);
		int op(const char*, uint32_t *val);
		int op(const char*, uint64_t *val);

		int op(const char*, bool *val);

		template<typename T>
		int op(const char*, T *val, std::size_t len);

		template<typename T>
		int op(const char*, ox::Vector<T> *val);

		template<std::size_t L>
		int op(const char*, const char *val);

		template<std::size_t L>
		int op(const char*, ox::BString<L> *val);

		template<typename T>
		int op(const char*, T *val);

		void setTypeInfo(const char *name, int fields);

		std::size_t size();

      OpType opType() {
          return OpType::Write;
      }

	private:
		template<typename I>
		int appendInteger(I val);
};

template<std::size_t L>
int MetalClawWriter::op(const char*, ox::BString<L> *val) {
	int err = 0;
	bool fieldSet = false;
	if (val->len()) {
		// write the length
		typedef uint32_t StringLength;
		if (m_buffIt + sizeof(StringLength) + val->bytes() < m_buffLen) {
			*reinterpret_cast<LittleEndian<StringLength>*>(&m_buff[m_buffIt]) = static_cast<StringLength>(val->bytes());
			m_buffIt += sizeof(StringLength);

			// write the string
			ox_memcpy(&m_buff[m_buffIt], val, val->bytes());
			m_buffIt += val->bytes();
			fieldSet = true;
		} else {
			err = MC_BUFFENDED;
		}
	}
	err |= m_fieldPresence.set(m_field, fieldSet);
	m_field++;
	return err;
}

template<typename T>
int MetalClawWriter::op(const char*, T *val) {
	int err = 0;
	bool fieldSet = false;
	MetalClawWriter writer(m_buff + m_buffIt, m_buffLen - m_buffIt);
	if (val) {
		err |= ioOp(&writer, val);
		if (static_cast<std::size_t>(writer.m_fieldPresence.getMaxLen()) < writer.m_buffIt) {
			m_buffIt += writer.m_buffIt;
			fieldSet = true;
		}
	}
	err |= m_fieldPresence.set(m_field, fieldSet);
	m_field++;
	return err;
}

template<typename T>
int MetalClawWriter::op(const char*, ox::Vector<T> *val) {
	return op(nullptr, val->data(), val->size());
}

template<typename I>
int MetalClawWriter::appendInteger(I val) {
	int err = 0;
	bool fieldSet = false;
	if (val) {
		if (m_buffIt + sizeof(I) < m_buffLen) {
			*reinterpret_cast<LittleEndian<I>*>(&m_buff[m_buffIt]) = val;
			fieldSet = true;
			m_buffIt += sizeof(I);
		} else {
			err |= MC_BUFFENDED;
		}
	}
	err |= m_fieldPresence.set(m_field, fieldSet);
	m_field++;
	return err;
}

template<typename T>
int MetalClawWriter::op(const char*, T *val, std::size_t len) {
	int err = 0;
	bool fieldSet = false;

	if (len) {
		// write the length
		typedef uint32_t ArrayLength;
		if (m_buffIt + sizeof(ArrayLength) < m_buffLen) {
			*reinterpret_cast<LittleEndian<ArrayLength>*>(&m_buff[m_buffIt]) = static_cast<ArrayLength>(len);
			m_buffIt += sizeof(ArrayLength);
		} else {
			err = MC_BUFFENDED;
		}

		MetalClawWriter writer(m_buff + m_buffIt, m_buffLen - m_buffIt);
		writer.setTypeInfo("List", len);

		// write the array
		for (std::size_t i = 0; i < len; i++) {
			err |= writer.op("", &val[i]);
		}

		m_buffIt += writer.m_buffIt;
		fieldSet = true;
	}

	err |= m_fieldPresence.set(m_field, fieldSet);
	m_field++;
	return err;
}

template<typename T>
int writeMC(uint8_t *buff, std::size_t buffLen, T *val, std::size_t *sizeOut = nullptr) {
	MetalClawWriter writer(buff, buffLen);
	auto err = ioOp(&writer, val);
	if (sizeOut) {
		*sizeOut = writer.size();
	}
	return err;
}

}
